@@ -53,6 +53,9 @@ struct _ModulePathElement {
 	char *method_name;
 	char *path;
 	char *args;
+
+	/* options */
+	gboolean daemon;
 };
 
 typedef struct _VfsDirSource VfsDirSource;
@@ -115,9 +118,7 @@ vfs_dir_source_free (VfsDirSource *vfs_source)
 
 
 static void
-hash_free_module_path (gpointer key,
-		       gpointer value,
-		       gpointer user_data)
+hash_free_module_path (gpointer value)
 {
 	ModulePathElement *module_path;
 
@@ -131,8 +132,6 @@ configuration_destroy (Configuration *configuration)
 {
 	g_return_if_fail (configuration != NULL);
 
-	g_hash_table_foreach (configuration->method_to_module_path,
-			      hash_free_module_path, NULL);
 	g_hash_table_destroy (configuration->method_to_module_path);
 	g_list_foreach (configuration->directories, (GFunc) vfs_dir_source_free, NULL);
 	g_list_free (configuration->directories);
@@ -215,9 +214,14 @@ parse_line (Configuration *configuration,
 	gchar *method_start;
 	char *module_name;
 	char *args = NULL;
+	char *option_start;
+	char *option;
 	GList *method_list;
 	GList *lp;
+	gboolean daemon;
 
+	daemon = FALSE;
+	
 	string_len = strlen (line_buffer);
 	if (string_len != line_len) {
 		g_warning (_("%s:%d contains NUL characters."),
@@ -264,6 +268,47 @@ parse_line (Configuration *configuration,
 	while (*p && g_ascii_isspace (*p))
 		p++;
 
+	if (*p == '[') {
+		p++;
+
+		while (TRUE) {
+			while (*p && g_ascii_isspace (*p))
+				p++;
+		
+			option_start = p;
+			while (*p && *p != ',' && *p != ']') {
+				p++;
+			}
+			
+			if (*p == '\0') {
+				g_warning (_("%s:%d has no options endmarker."),
+						   file_name, line_number);
+				retval = FALSE;
+				goto cleanup;
+			}
+
+			if (p != option_start) {
+				option = g_strndup (option_start, p - option_start);
+				if (strcmp (option, "daemon") == 0) {
+					daemon = TRUE;
+				} else {
+					g_warning (_("%s:%d has unknown options %s."),
+						   file_name, line_number, option);
+				}
+				g_free (option);
+			}
+
+			if (*p == ']') {
+				p++;
+				break;
+			}
+			p++;
+		}
+	}
+	
+	while (*p && g_ascii_isspace (*p))
+		p++;
+	
 	if (*p == '\0') {
 		if (method_list != NULL) {
 			g_warning (_("%s:%d contains no module name."),
@@ -293,6 +338,7 @@ parse_line (Configuration *configuration,
 
 		method_name = lp->data;
 		element = module_path_element_new (method_name, module_name, args);
+		element->daemon = daemon;
 		g_hash_table_insert (configuration->method_to_module_path,
 				     method_name, element);
 	}
@@ -355,7 +401,7 @@ configuration_load (void)
 	int i = 0;
 	DIR *dirh;
 
-	configuration->method_to_module_path = g_hash_table_new (g_str_hash, g_str_equal);
+	configuration->method_to_module_path = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, hash_free_module_path);
 
 	/* Go through the list of configuration directories and build up a list of config files */
 	for (list = configuration->directories; list && i < MAX_CFG_FILES; list = list->next) {
@@ -399,11 +445,11 @@ add_directory_internal (const char *dir)
 }
 
 void
-gnome_vfs_configuration_add_directory (const char *dir)
+_gnome_vfs_configuration_add_directory (const char *dir)
 {
 	G_LOCK (configuration);
 	if (configuration == NULL) {
-		g_warning ("gnome_vfs_configuration_init must be called prior to adding a directory.");
+		g_warning ("_gnome_vfs_configuration_init must be called prior to adding a directory.");
 		G_UNLOCK (configuration);
 		return;
 	}
@@ -442,10 +488,11 @@ install_path_list (const gchar *environment_path)
 
 
 gboolean
-gnome_vfs_configuration_init (void)
+_gnome_vfs_configuration_init (void)
 {
 	char *home_config;
 	char *environment_path;
+	const char *home_dir;
 
 	G_LOCK (configuration);
 	if (configuration != NULL) {
@@ -455,30 +502,35 @@ gnome_vfs_configuration_init (void)
 
 	configuration = g_new0 (Configuration, 1);
 
-	home_config = g_strdup_printf ("%s%c%s",
-				       g_get_home_dir (),
-				       G_DIR_SEPARATOR,
-				       ".gnome/vfs/modules");
 	add_directory_internal (GNOME_VFS_MODULE_CFGDIR);
 	environment_path = getenv ("GNOME_VFS_MODULE_CONFIG_PATH");
 	if (environment_path != NULL) {
 		install_path_list (environment_path);
 	}
-	add_directory_internal (home_config);
-	g_free (home_config);
+
+	home_dir = g_get_home_dir ();
+	if (home_dir != NULL) {
+		home_config = g_strdup_printf ("%s%c%s",
+					       home_dir,
+					       G_DIR_SEPARATOR,
+					       ".gnome2/vfs/modules");
+		add_directory_internal (home_config);
+		g_free (home_config);
+	}
 
 	configuration_load ();
 
 	G_UNLOCK (configuration);
 
-	if (configuration == NULL)
+	if (configuration == NULL) {
 		return FALSE;
-	else
+	} else {
 		return TRUE;
+	}
 }
 
 void
-gnome_vfs_configuration_uninit (void)
+_gnome_vfs_configuration_uninit (void)
 {
 	G_LOCK (configuration);
 	if (configuration == NULL) {
@@ -520,14 +572,12 @@ maybe_reload (void)
 
 	configuration->last_checked = time (NULL);
 
-	g_hash_table_foreach (configuration->method_to_module_path,
-			      hash_free_module_path, NULL);
 	g_hash_table_destroy (configuration->method_to_module_path);
 	configuration_load ();
 }
 
 const gchar *
-gnome_vfs_configuration_get_module_path (const gchar *method_name, const char ** args)
+_gnome_vfs_configuration_get_module_path (const gchar *method_name, const char ** args, gboolean *daemon)
 {
 	ModulePathElement *element;
 
@@ -541,7 +591,7 @@ gnome_vfs_configuration_get_module_path (const gchar *method_name, const char **
 			(configuration->method_to_module_path, method_name);
 	} else {
 		/* This should never happen.  */
-		g_warning ("Internal error: the configuration system was not initialized. Did you call gnome_vfs_configuration_init?");
+		g_warning ("Internal error: the configuration system was not initialized. Did you call _gnome_vfs_configuration_init?");
 		element = NULL;
 	}
 
@@ -552,5 +602,32 @@ gnome_vfs_configuration_get_module_path (const gchar *method_name, const char **
 
 	if (args)
 		*args = element->args;
+	if (daemon)
+		*daemon = element->daemon;
 	return element->path;
 }
+
+static void
+add_method_to_list(const gchar *key, gpointer value, GList **methods_list)
+{
+	*methods_list = g_list_append(*methods_list, g_strdup(key));
+}
+
+GList *
+_gnome_vfs_configuration_get_methods_list (void)
+{
+	GList *methods_list = NULL;
+
+	G_LOCK (configuration);
+	if (configuration != NULL) {
+		maybe_reload ();
+		g_hash_table_foreach(configuration->method_to_module_path, 
+				     (GHFunc)add_method_to_list, &methods_list);
+	} else {
+		/* This should never happen.  */
+		methods_list = NULL;
+	}
+
+	G_UNLOCK (configuration);
+	return methods_list;
+}
