@@ -33,8 +33,8 @@ System (version for POSIX threads).
 #include "gnome-vfs-async-job-map.h"
 #include "gnome-vfs-job-slave.h"
 #include "gnome-vfs-job-queue.h"
+#include "gnome-vfs-private-utils.h"
 #include <errno.h>
-#include <fcntl.h>
 #include <glib/gmessages.h>
 #include <glib/gstrfuncs.h>
 #include <libgnomevfs/gnome-vfs-cancellable-ops.h>
@@ -44,7 +44,7 @@ System (version for POSIX threads).
 #include <string.h>
 #include <unistd.h>
 
-GStaticPrivate job_private = G_STATIC_PRIVATE_INIT;
+static GStaticPrivate job_private = G_STATIC_PRIVATE_INIT;
 
 #if GNOME_VFS_JOB_DEBUG
 
@@ -52,7 +52,7 @@ char *job_debug_types[] = {
 	"open", "open as channel",
 	"create", "create symbolic link",
 	"create as channel", "close",
-	"read", "write", "read write done",
+	"read", "write", "seek", "read write done",
 	"load directory", "find directory",
 	"xfer", "get file info", "set file info",
 	"module callback", "file control",
@@ -78,46 +78,6 @@ static gboolean dispatch_sync_job_callback          (gpointer              data)
 static void	clear_current_job 		    (void);
 static void	set_current_job 		    (GnomeVFSJob *context);
 
-static void
-set_fl (int fd, int flags)
-{
-	int val;
-
-	val = fcntl (fd, F_GETFL, 0);
-	if (val < 0) {
-		g_warning ("fcntl() F_GETFL failed: %s", strerror (errno));
-		return;
-	}
-
-	val |= flags;
-	
-	val = fcntl (fd, F_SETFL, val);
-	if (val < 0) {
-		g_warning ("fcntl() F_SETFL failed: %s", strerror (errno));
-		return;
-	}
-}
-
-static void
-clr_fl (int fd, int flags)
-{
-	int val;
-
-	val = fcntl (fd, F_GETFL, 0);
-	if (val < 0) {
-		g_warning ("fcntl() F_GETFL failed: %s", strerror (errno));
-		return;
-	}
-
-	val &= ~flags;
-	
-	val = fcntl (fd, F_SETFL, val);
-	if (val < 0) {
-		g_warning ("fcntl() F_SETFL failed: %s", strerror (errno));
-		return;
-	}
-}
-
 /*
  *   Find out whether or not a given job should be left in
  * the job map, preserving it's open VFS handle, since we
@@ -143,6 +103,7 @@ _gnome_vfs_job_complete (GnomeVFSJob *job)
 		return FALSE;
 	case GNOME_VFS_OP_READ_WRITE_DONE:
 	case GNOME_VFS_OP_FILE_CONTROL:
+	case GNOME_VFS_OP_SEEK:
 		return FALSE;
 	
 	default:
@@ -167,6 +128,8 @@ job_oneway_notify (GnomeVFSJob *job, GnomeVFSNotifyResult *notify_result)
 		JOB_DEBUG (("Barfing on oneway cancel %u (%d) type '%s'",
 			    GPOINTER_TO_UINT (notify_result->job_handle),
 			    job->op->type, JOB_DEBUG_TYPE (job->op->type)));
+		/* TODO: We can leak handle here, if an open succeded.
+		 * See bug #123472 */
 		_gnome_vfs_job_destroy_notify_result (notify_result);
 	}
 }
@@ -260,6 +223,14 @@ dispatch_write_callback (GnomeVFSNotifyResult *notify_result)
 						     notify_result->specifics.write.callback_data);
 }
 
+static void
+dispatch_seek_callback (GnomeVFSNotifyResult *notify_result)
+{
+	(* notify_result->specifics.seek.callback) (notify_result->job_handle,
+						    notify_result->specifics.seek.result,
+						    notify_result->specifics.seek.callback_data);
+}
+
 static void
 dispatch_load_directory_callback (GnomeVFSNotifyResult *notify_result)
 {
@@ -396,6 +367,7 @@ _gnome_vfs_job_destroy_notify_result (GnomeVFSNotifyResult *notify_result)
 	case GNOME_VFS_OP_CREATE_AS_CHANNEL:
 	case GNOME_VFS_OP_CREATE_SYMBOLIC_LINK:
 	case GNOME_VFS_OP_WRITE:
+	case GNOME_VFS_OP_SEEK:
 	case GNOME_VFS_OP_OPEN:
 	case GNOME_VFS_OP_OPEN_AS_CHANNEL:
 	case GNOME_VFS_OP_READ:
@@ -604,6 +576,9 @@ dispatch_job_callback (gpointer data)
 	case GNOME_VFS_OP_WRITE:
 		dispatch_write_callback (notify_result);
 		break;
+	case GNOME_VFS_OP_SEEK:
+		dispatch_seek_callback (notify_result);
+		break;
 	case GNOME_VFS_OP_FILE_CONTROL:
 		dispatch_file_control_callback (notify_result);
 		break;
@@ -748,6 +723,7 @@ gnome_vfs_op_destroy (GnomeVFSOp *op)
 		break;
 	case GNOME_VFS_OP_READ:
 	case GNOME_VFS_OP_WRITE:
+	case GNOME_VFS_OP_SEEK:
 	case GNOME_VFS_OP_CLOSE:
 	case GNOME_VFS_OP_READ_WRITE_DONE:
 		break;
@@ -755,7 +731,7 @@ gnome_vfs_op_destroy (GnomeVFSOp *op)
 		g_free (op->specifics.file_control.operation);
 		break;
 	default:
-		g_warning (_("Unknown op type %u"), op->type);
+		g_warning (_("Unknown operation type %u"), op->type);
 	}
 	
 	g_assert (gnome_vfs_context_get_cancellation (op->context) != NULL);
@@ -871,7 +847,7 @@ serve_channel_read (GnomeVFSHandle *handle,
 
 					fd = g_io_channel_unix_get_fd (channel_out);
 					
-					clr_fl (fd, O_NONBLOCK);
+					_gnome_vfs_clear_fd_flags (fd, O_NONBLOCK);
 				} else {
 					if (written_bytes_in_buffer > 0) {
 						/* Need to shift the unwritten bytes
@@ -1058,7 +1034,7 @@ execute_open_as_channel (GnomeVFSJob *job)
 	 * thread is blocking for some reason the slave can keep
 	 * reading data.
 	 */
-	set_fl (pipefd[1], O_NONBLOCK);
+	_gnome_vfs_set_fd_flags (pipefd[1], O_NONBLOCK);
 	
 	channel_in = g_io_channel_unix_new (pipefd[0]);
 	channel_out = g_io_channel_unix_new (pipefd[1]);
@@ -1282,6 +1258,30 @@ execute_write (GnomeVFSJob *job)
 	job_oneway_notify (job, notify_result);
 }
 
+static void
+execute_seek (GnomeVFSJob *job)
+{
+	GnomeVFSResult result;
+	GnomeVFSSeekOp *seek_op;
+	GnomeVFSNotifyResult *notify_result;
+
+	seek_op = &job->op->specifics.seek;
+
+	result = gnome_vfs_seek_cancellable (job->handle,
+					     seek_op->whence,
+					     seek_op->offset,
+					     job->op->context);
+
+	notify_result = g_new0 (GnomeVFSNotifyResult, 1);
+	notify_result->job_handle = job->job_handle;
+	notify_result->type = job->op->type;
+	notify_result->specifics.seek.result = result;
+	notify_result->specifics.seek.callback = (GnomeVFSAsyncSeekCallback) job->op->callback;
+	notify_result->specifics.seek.callback_data = job->op->callback_data;
+
+	job_oneway_notify (job, notify_result);
+}
+
 static void
 execute_get_file_info (GnomeVFSJob *job)
 {
@@ -1662,6 +1662,9 @@ _gnome_vfs_job_execute (GnomeVFSJob *job)
 		case GNOME_VFS_OP_WRITE:
 			execute_write (job);
 			break;
+		case GNOME_VFS_OP_SEEK:
+			execute_seek (job);
+			break;
 		case GNOME_VFS_OP_LOAD_DIRECTORY:
 			execute_load_directory (job);
 			break;
