@@ -36,6 +36,7 @@
 #include "gnome-vfs-private-utils.h"
 #include "gnome-vfs-ops.h"
 #include "gnome-vfs-mime-handlers.h"
+#include "gnome-vfs-mime-private.h"
 #include <glib/gstrfuncs.h>
 #include <glib/gutils.h>
 #include <gconf/gconf-client.h>
@@ -81,7 +82,7 @@ gchar*
 gnome_vfs_format_file_size_for_display (GnomeVFSFileSize size)
 {
 	if (size < (GnomeVFSFileSize) KILOBYTE_FACTOR) {
-		return g_strdup_printf (ngettext("%u byte", "%u bytes",(guint) size), (guint) size);
+		return g_strdup_printf (dngettext(GETTEXT_PACKAGE, "%u byte", "%u bytes",(guint) size), (guint) size);
 	} else {
 		gdouble displayed_size;
 
@@ -639,7 +640,7 @@ gnome_vfs_make_uri_canonical_old (const char *original_uri_text)
 
 	uri = gnome_vfs_uri_new_private (original_uri_text, TRUE, TRUE, FALSE);
 	if (uri == NULL) {
-		return NULL;;
+		return NULL;
 	} 
 
 	result = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE);
@@ -796,8 +797,15 @@ gnome_vfs_get_volume_free_space (const GnomeVFSURI *vfs_uri,
 	
 #if HAVE_STATVFS
 	statfs_result = statvfs (unescaped_path, &statfs_buffer);
+	block_size = statfs_buffer.f_frsize; 
 #else
-	statfs_result = statfs (unescaped_path, &statfs_buffer);   
+#if STATFS_ARGS == 2
+	statfs_result = statfs (unescaped_path, &statfs_buffer);
+#elif STATFS_ARGS == 4
+	statfs_result = statfs (unescaped_path, &statfs_buffer,
+				sizeof (statfs_buffer), 0);
+#endif
+	block_size = statfs_buffer.f_bsize; 
 #endif  
 
 	if (statfs_result != 0) {
@@ -805,9 +813,14 @@ gnome_vfs_get_volume_free_space (const GnomeVFSURI *vfs_uri,
 		return gnome_vfs_result_from_errno ();
 	}
 
+
+/* CF: I assume ncpfs is linux specific, if you are on a non-linux platform
+ * where ncpfs is available, please file a bug about it on bugzilla.gnome.org
+ * (2004-03-08)
+ */
+#if defined(__linux__)
 	/* ncpfs does not know the amount of available and free space */
 	if (statfs_buffer.f_bavail == 0 && statfs_buffer.f_bfree == 0) {
-#if defined(__linux__)
 		/* statvfs does not contain an f_type field, we try again
 		 * with statfs.
 		 */
@@ -821,15 +834,16 @@ gnome_vfs_get_volume_free_space (const GnomeVFSURI *vfs_uri,
 		
 		/* linux/ncp_fs.h: NCP_SUPER_MAGIC == 0x564c */
 		if (statfs_buffer2.f_type == 0x564c)
-#elif defined(__sun)
-		if (strcmp(statfs_buffer.f_basetype, "ncpfs") == 0)
-#endif
 		{
 			return GNOME_VFS_ERROR_NOT_SUPPORTED;
 		}
+	} else {
+		/* everything is copacetic... free the unescaped path */
+		g_free (unescaped_path);
 	}
-
-	block_size = statfs_buffer.f_bsize; 
+#else 	
+	g_free (unescaped_path);
+#endif
 	free_blocks = statfs_buffer.f_bavail;
 
 	*size = block_size * free_blocks;
@@ -1023,6 +1037,9 @@ gnome_vfs_read_entire_file (const char *uri,
 	GnomeVFSFileSize total_bytes_read;
 	GnomeVFSFileSize bytes_read;
 
+	g_assert (file_size != NULL);
+	g_assert (file_contents != NULL);
+
 	*file_size = 0;
 	*file_contents = NULL;
 
@@ -1169,7 +1186,7 @@ gnome_vfs_format_uri_for_display (const char *uri)
 {
 	static gboolean broken_filenames;
 	
-	broken_filenames = g_getenv ("G_BROKEN_FILENAMES") != NULL;
+	broken_filenames = (g_getenv ("G_BROKEN_FILENAMES") != NULL);
 
 	return gnome_vfs_format_uri_for_display_internal (uri, broken_filenames);
 }
@@ -1355,7 +1372,7 @@ gnome_vfs_make_uri_from_input_internal (const char *text,
 
 /**
  * gnome_vfs_make_uri_from_input:
- * @location: a possibly mangled "uri", in UTF8
+ * @uri: a possibly mangled "uri", in UTF8
  *
  * Takes a user input path/URI and makes a valid URI out of it.
  *
@@ -1372,14 +1389,15 @@ gnome_vfs_make_uri_from_input (const char *location)
 {
 	static gboolean broken_filenames;
 
-	broken_filenames = g_getenv ("G_BROKEN_FILENAMES") != NULL;
+	broken_filenames = (g_getenv ("G_BROKEN_FILENAMES") != NULL);
 
 	return gnome_vfs_make_uri_from_input_internal (location, broken_filenames, TRUE);
 }
 
 /**
  * gnome_vfs_make_uri_from_input_with_dirs:
- * @in: a relative or absolute path
+ * @uri: a relative or absolute path
+ * @dirs: directory to use as a base directory if @uri is a relative path.
  *
  * Determines a fully qualified URL from a relative or absolute input path.
  * Basically calls gnome_vfs_make_uri_from_input except it specifically
@@ -1740,8 +1758,8 @@ gnome_vfs_get_uri_scheme (const char *uri)
 	if (colon == NULL) {
 		return NULL;
 	}
-	
-	return g_strndup (uri, colon - uri);
+
+	return g_ascii_strdown (uri, colon - uri);
 }
 
 /* Note that NULL's and full paths are also handled by this function.
@@ -1774,7 +1792,7 @@ file_uri_from_local_relative_path (const char *location)
 
 /**
  * gnome_vfs_make_uri_from_shell_arg:
- * @location: a possibly mangled "uri"
+ * @uri: a possibly mangled "uri"
  *
  * Similar to gnome_vfs_make_uri_from_input, except that:
  * 
@@ -1814,7 +1832,9 @@ gnome_vfs_make_uri_from_shell_arg (const char *location)
 
 /**
  * gnome_vfs_make_uri_full_from_relative:
- * 
+ * @base_uri: uri to use as the base for the full uri
+ * @relative_uri: full or relative path to append to the base uri
+ *
  * Returns a full URI given a full base URI, and a secondary URI which may
  * be relative.
  *
@@ -1827,7 +1847,8 @@ gnome_vfs_make_uri_from_shell_arg (const char *location)
  **/
 
 char *
-gnome_vfs_make_uri_full_from_relative (const char *base_uri, const char *relative_uri)
+gnome_vfs_make_uri_full_from_relative (const char *base_uri, 
+				       const char *relative_uri)
 {
 	return gnome_vfs_uri_make_full_from_relative (base_uri, relative_uri);
 }
@@ -1909,7 +1930,12 @@ _gnome_vfs_uri_resolve_all_symlinks (const char *text_uri,
 	*resolved_text_uri = NULL;
 
 	uri = gnome_vfs_uri_new (text_uri);
-	if (uri == NULL || uri->text == NULL) {
+
+	if (uri == NULL) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+	if (uri->text == NULL) {
+		gnome_vfs_uri_unref (uri);
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
 
@@ -1919,6 +1945,7 @@ _gnome_vfs_uri_resolve_all_symlinks (const char *text_uri,
 		*resolved_text_uri = gnome_vfs_uri_to_string (resolved_uri, GNOME_VFS_URI_HIDE_NONE);
 		gnome_vfs_uri_unref (resolved_uri);
 	}
+	gnome_vfs_uri_unref (uri);
 	return res;
 }
 
@@ -2049,11 +2076,10 @@ gnome_vfs_url_show_with_env (const char  *url,
 		return result;
 	}
 	
-	g_free (scheme);
-
-	type = gnome_vfs_get_mime_type (url);
+	type = _gnome_vfs_get_slow_mime_type (url);
 
 	if (type == NULL) {
+		g_free (scheme);
 		return GNOME_VFS_ERROR_BAD_PARAMETERS;
 	}
 
@@ -2061,7 +2087,8 @@ gnome_vfs_url_show_with_env (const char  *url,
 	params.prev = NULL;
 	params.next = NULL;
 	
-	app = gnome_vfs_mime_get_default_application (type);
+	app = gnome_vfs_mime_get_default_application_for_scheme (type, scheme);
+	g_free (scheme);
 	
 	if (app != NULL) {
 		result = gnome_vfs_mime_application_launch_with_env (app, &params, envp);
